#!/usr/bin/env python
"""
DAFoam run script for the topology optimization of the pitzDaily case
"""

# =============================================================================
# Imports
# =============================================================================
import os
import argparse
from mpi4py import MPI
from dafoam import PYDAFOAM, optFuncs
from pygeo import *
from pyspline import *
from idwarp import USMesh
from pyoptsparse import Optimization, OPT
import numpy as np


# =============================================================================
# Input Parameters
# =============================================================================
parser = argparse.ArgumentParser()
parser.add_argument("--opt", help="optimizer to use", type=str, default="snopt")
parser.add_argument("--task", help="type of run to do", type=str, default="sens")
args = parser.parse_args()
gcomm = MPI.COMM_WORLD

# Define the global parameters here
U0 = 40.0
p0 = 0.0
nuTilda0 = 1.0e-4
nCells = 12225

# Set the parameters for optimization
daOptions = {
    "designSurfaces": ["upperWall"],
    "solverName": "DASimpleFoam",
    "adjJacobianOption": "JacobianFree",
    "primalMinResTol": 1.0e-6,
    "primalMinResTolDiff": 1.0e4,
    "primalBC": {
        "U0": {"variable": "U", "patches": ["inlet"], "value": [U0, 0.0, 0.0]},
        "p0": {"variable": "p", "patches": ["outlet"], "value": [p0]},
        "nuTilda0": {"variable": "nuTilda", "patches": ["inlet"], "value": [nuTilda0]},
        "useWallFunction": True,
    },
    "objFunc": {
         "PL": {
            "part1": {
                "type": "totalPressure",
                "source": "patchToFace",
                "patches": ["inlet"],
                "scale": 1.0 / (0.5 * U0 * U0),
                "addToAdjoint": True,
            },
            "part2": {
                "type": "totalPressure",
                "source": "patchToFace",
                "patches": ["outlet"],
                "scale": -1.0 / (0.5 * U0 * U0),
                "addToAdjoint": True,
            }
        }
    },
    "adjEqnOption": {"gmresRelTol": 1.0e-6, "gmresMaxIters": 2000, "gmresRestart": 2000, "pcFillLevel": 1, "jacMatReOrdering": "rcm"},
    "normalizeStates": {
        "U": U0,
        "p": U0 * U0 / 2.0,
        "nut": nuTilda0 * 10.0,
        "phi": 1.0,
    },
    "adjPCLag": 5,
    "designVar": {},
}

# mesh warping parameters, users need to manually specify the symmetry plane and their normals
meshOptions = {
    "gridFile": os.getcwd(),
    "fileType": "openfoam",
    # point and normal for the symmetry plane
    "symmetryPlanes": [[[0.0, 0.0, -0.0005], [0.0, 0.0, 1.0]], [[0.0, 0.0, 0.0005], [0.0, 0.0, 1.0]]],
}

# options for optimizers
if args.opt == "snopt":
    optOptions = {
        "Major feasibility tolerance": 1.0e-7,
        "Major optimality tolerance": 1.0e-7,
        "Minor feasibility tolerance": 1.0e-7,
        "Verify level": -1,
        "Function precision": 1.0e-5,
        "Hessian updates": 20,
        "Nonderivative linesearch": None,
        "Print file": "opt_SNOPT_print.txt",
        "Summary file": "opt_SNOPT_summary.txt",
    }
elif args.opt == "ipopt":
    optOptions = {
        "tol": 1.0e-7,
        "constr_viol_tol": 1.0e-7,
        "max_iter": 50,
        "print_level": 5,
        "output_file": "opt_IPOPT.txt",
        "mu_strategy": "adaptive",
        "limited_memory_max_history": 25,
        "nlp_scaling_method": "none",
        "alpha_for_y": "full",
        "recalc_y": "yes",
    }
elif args.opt == "slsqp":
    optOptions = {
        "ACC": 1.0e-7,
        "MAXIT": 50,
        "IFILE": "opt_SLSQP.txt",
    }
else:
    print("opt arg not valid!")
    exit(0)


# =============================================================================
# Design variable setup
# =============================================================================
def alphaPorosity(val, geo):
    for idxI, v in enumerate(val):
        DASolver.setFieldValue4GlobalCellI(b"alphaPorosity", v, idxI)
        DASolver.updateBoundaryConditions(b"alphaPorosity", b"scalar")

DVGeo = DVGeometry("./FFD/dummyFFD.xyz")
# We need to add a dummy ref axis for defining GeoDVGlobal
DVGeo.addRefAxis("dummyAxis", xFraction=0.25, alignIndex="k")

alphaPorosity0 = np.zeros(nCells, dtype="d")
# modify one alpha value. This can be used to verify the adjoint derivative with the finite-difference method
alphaPorosity0[4521] = 1e-5
DVGeo.addGeoDVGlobal("alphaPorosity", value=alphaPorosity0, func=alphaPorosity, lower=0, upper=1e4, scale=1e-4)
daOptions["designVar"]["alphaPorosity"] = {"designVarType": "Field", "fieldName": "alphaPorosity", "fieldType": "scalar"}

# =============================================================================
# DAFoam initialization
# =============================================================================
DASolver = PYDAFOAM(options=daOptions, comm=gcomm)
DASolver.setDVGeo(DVGeo)
mesh = USMesh(options=meshOptions, comm=gcomm)
DASolver.addFamilyGroup(DASolver.getOption("designSurfaceFamily"), DASolver.getOption("designSurfaces"))
DASolver.printFamilyList()
DASolver.setMesh(mesh)
evalFuncs = []
DASolver.setEvalFuncs(evalFuncs)

# =============================================================================
# Constraint setup
# =============================================================================
DVCon = DVConstraints()
DVCon.setDVGeo(DVGeo)
DVCon.setSurface(DASolver.getTriangulatedMeshSurface(groupName=DASolver.getOption("designSurfaceFamily")))
# =============================================================================
# Initialize optFuncs for optimization
# =============================================================================
optFuncs.DASolver = DASolver
optFuncs.DVGeo = DVGeo
optFuncs.DVCon = DVCon
optFuncs.evalFuncs = evalFuncs
optFuncs.gcomm = gcomm


def write_sens(sensitivity, filename):
    nvals = len(sensitivity)

    with open(filename, "w") as ofile:
        s = '/*--------------------------------*- C++ -*----------------------------------*\n' \
            '| =========                 |                                                 |\n' \
            '| \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox           |\n' \
            '|  \\    /   O peration     | Version:  v2006                                 |\n' \
            '|   \\  /    A nd           | Website:  www.openfoam.com                      |\n' \
            '|    \\/     M anipulation  |                                                 |\n' \
            '\*---------------------------------------------------------------------------*/\n' \
            'FoamFile\n' \
            '{\n' \
            '    version     2.0;\n' \
            '    format      ascii;\n' \
            '    class       volScalarField;\n' \
            '    location    "0";\n' \
            '    object      sens;\n' \
            '}\n' \
            '// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n' \
            '\n' \
            'dimensions      [0 0 0 0 0 0 0];\n' \
            'internalField   nonuniform List<scalar>\n'
        s += f"{nvals}\n"
        s += "(\n"
        for val in sensitivity:
            s += f"{val}\n"
        s += ")"
        ofile.write(s)


# =============================================================================
# Task
# =============================================================================
if args.task == "sens":

    DASolver.runColoring()
    xDV = DVGeo.getValues()
    funcs, fail = optFuncs.calcObjFuncValues(xDV)
    sens, fail = optFuncs.calcObjFuncSens(xDV, funcs)
    write_sens(sensitivity=sens["PL"]["alphaPorosity"], filename=os.path.join("0", "sensitivity"))

else:
    print("task arg not found!")
    exit(0)
